#include <engine/allocmem.h>
#include <engine/controls.h>
#include <engine/display.h>
#include <engine/math.h>
#include <engine/music.h>
#include <engine/sfx.h>
#include <engine/3d/render.h>
#include <engine/2d/level.h>
#include <game/settings.h>

void
init_graphics(void)
{
	int i, j;
	RECT rect;

	SetVideoMode(REGION);
	ResetGraph(0);

	/* clear VRAM */
	setRECT(&rect, 0, 0, 1024, 512);
	ClearImage(&rect, 0, 0, 0);
	DrawSync(0);

	/* load built-in font */
	FntLoad(960, 0);
	SetDumpFnt(FntOpen(5, 20, screen_width, screen_height, 0, 512));

	/* initialize on-screen section's tileset */
	{
		unsigned int row = 0, col = 0;

		/* initialize on-screen section tileset */
		for(i = 1; i < SCT_UNIQUE_TILES; i++)
		{
			sct.tileset[i].u = col * (SCT_TILE_SPRITE_SIZE + SCT_TILE_SPACING);
			sct.tileset[i].v = row * (SCT_TILE_SPRITE_SIZE + SCT_TILE_SPACING);

			col++;

			/* if the next tile exceeds the texture page's width, do a CRLF */
			if( !(i % (256/(SCT_TILE_SPRITE_SIZE + SCT_TILE_SPACING))) )
			{
				row++;
				col = 0;
			}
		}
	}

	/* initialize light sources */
	for(i = 0; i < 3; i++)
	{
		setVector(dgx_light_sun_ang+i, 0, 0, 0);
		setVector(dgx_light_point_pos+i, 0, 0, 0);

		for(j = 0; j < 3; j++)
		{
			dgx_light_sun_color.m[i][j] = 0;
			dgx_light_sun.m[i][j] = 0;
		}
	}
	dgx_light_flag = 0;

	/* initialize subdivision buffers */
	divbuf_init();

	game_otsize = GAME_OTSIZE_MAX;

	VSyncCallback(cb_vsync);

	/* enable_display */
	SetDispMask(1);
}

void
init_sound_soft(void)
{
	int i;
	CdInit();
	CdSetDebug(0);
	cdsetup_mode(CdlModeRT|CdlModeSF);
	CdReadyCallback((CdlCB)cbready);

	SpuSetCommonAttr(&game_settings.cattr);
	for(i = 0; i < 16; i++)
	{
		SpuNSetVoiceAttr(i, game_settings.sattr+i);
	}
}

void
init_sound(void)
{
	int i;

#ifndef EXE_BUILD
	CdInit();
	CdSetDebug(0);

	/* initialize music XA file */
	if(!cdsetup_XA(&music, "\\MUSIC.XA;1", 0, 1))
	{
		/* must restart console at this point */
		while(1)
		{
			switch_double_buffer();

			FntPrint("FAILED TO LOAD %s!\n", "\\MUSIC.XA");
			FntFlush(-1);

			DrawSync(0);
			VSync(0);
		}
	}

	cdsetup_mode(CdlModeSpeed|CdlModeRT|CdlModeSF);
	//cdsetup_mode(CdlModeRT|CdlModeSF);
	//CdReadyCallback((CdlCB)cbready);

	/* clean data buffer */
	for(i = 0; i < DATABUFFER_SIZE; i++)
	{
		data_buffer[i] = 0;
	}

	game_settings.cdlatv.val0 = game_settings.cdlatv.val2 = 0x80;
	game_settings.cdlatv.val1 = game_settings.cdlatv.val3 = 0x80;
	CdMix(&game_settings.cdlatv);
#endif

	///* initialize SPU */
	SpuInit();
	SpuInitMalloc(SPU_MALLOC_MAX, spu_malloc_rec);

	/* set common attributes */
	game_settings.cattr.mask = (SPU_COMMON_CDVOLL|SPU_COMMON_CDVOLR);
	game_settings.cattr.cd.mix = SPU_ON;
	game_settings.cattr.cd.reverb = SPU_ON;
	game_settings.cattr.cd.volume.left = 0x3fff;
	game_settings.cattr.cd.volume.right = 0x3fff;

	//cattr.mvol.left = 0x3fff;
	//cattr.mvol.right = 0x3fff;
	SpuSetCommonAttr(&game_settings.cattr);

	SpuSetTransferMode(SpuTransByDMA);
	for(i = 0; i < NSFX; i++)
	{
		/* set voice attributes */
		game_settings.sattr[i].voice = 1<<i;
		game_settings.sattr[i].mask = (SPU_VOICE_VOLL |
			SPU_VOICE_VOLR |
			SPU_VOICE_PITCH |
			SPU_VOICE_WDSA |
			SPU_VOICE_ADSR_AMODE |
			SPU_VOICE_ADSR_SMODE |
			SPU_VOICE_ADSR_RMODE |
			SPU_VOICE_ADSR_AR |
			SPU_VOICE_ADSR_DR |
			SPU_VOICE_ADSR_SR |
			SPU_VOICE_ADSR_RR |
			SPU_VOICE_ADSR_SL
			);

		game_settings.sattr[i].volume.left	= 0x1fff;
		game_settings.sattr[i].volume.right	= 0x1fff;
		game_settings.sattr[i].pitch		= 0x1000;
		game_settings.sattr[i].a_mode		= SPU_VOICE_LINEARIncN;
		game_settings.sattr[i].s_mode		= SPU_VOICE_LINEARIncN;
		game_settings.sattr[i].r_mode		= SPU_VOICE_LINEARDecN;
		game_settings.sattr[i].ar		= 0x0;
		game_settings.sattr[i].dr		= 0x0;
		game_settings.sattr[i].sr		= 0x0;
		game_settings.sattr[i].rr		= 0x0;
		game_settings.sattr[i].sl		= 0xf;
	}
}

void
init_double_buffer(screen_width, screen_height)
int screen_width, screen_height;
{
	SetDefDrawEnv(&db[0].draw, 0,             0, screen_width, screen_height); /* draw[0] -- top */
	SetDefDispEnv(&db[0].disp, 0, screen_height, screen_width, screen_height); /* disp[0] -- bottom */
	SetDefDrawEnv(&db[1].draw, 0, screen_height, screen_width, screen_height); /* draw[1] -- bottom */
	SetDefDispEnv(&db[1].disp, 0,             0, screen_width, screen_height); /* disp[1] -- top */

	/* clear BG with background color */
	db[0].draw.isbg = db[1].draw.isbg = 1;
	setRGB0(&db[0].draw, BG_RED, BG_GREEN, BG_BLUE);
	setRGB0(&db[1].draw, BG_RED, BG_GREEN, BG_BLUE);

#ifdef IS_INTERLACE
	SetDefDrawEnv(&db[0].draw, 0, 0, screen_width, screen_height);
	SetDefDispEnv(&db[0].disp, 0, 0, screen_width, screen_height);
	SetDefDrawEnv(&db[1].draw, 0, 0, screen_width, screen_height);
	SetDefDispEnv(&db[1].disp, 0, 0, screen_width, screen_height);
	db[0].disp.isinter = db[1].disp.isinter = 1;
#endif
}

void
init_controllers(void)
{
	InitPAD(padbuf[0], 34, padbuf[1], 34);
	StartPAD();

	gamepad[0].btn = gamepad[1].btn = 0;
	gamepad[0].btnprev = gamepad[1].btnprev = 0;

	gamepad[0].rx = gamepad[0].ry = gamepad[0].lx = gamepad[0].ly = 0;
	gamepad[1].rx = gamepad[1].ry = gamepad[1].lx = gamepad[1].ly = 0;
}

void
init_engine(void)
{
	screen_width = SCREEN_WIDTH_NORMAL;

#if REGION == NTSC_DISC
	screen_height = SCREEN_HEIGHT_NTSC;
#else
	screen_height = SCREEN_HEIGHT_PAL;
#endif

#ifdef IS_INTERLACE
	screen_width = SCREEN_WIDTH_INTERLACE;
	screen_height = SCREEN_HEIGHT_INTERLACE;
#endif

	ResetCallback();
	srand(FIXED_PRECISION);

	allocmem_clear(ALLOCBUF);
	allocmem_clear(ALLOCBUF_DC);

	/* initialize graphics & controller */
	init_graphics();
	init_double_buffer(screen_width, screen_height);

	init_controllers();

	/* initialize the GTE */
	InitGeom();
	SetGeomOffset(screen_width/2, screen_height/2);
	SetGeomScreen(DEFAULT_FOV);

	/* initialize the sound system */
	init_sound();

	/* initialize root counter */
	SetRCnt(RCntCNT2, UINT_MAX, RCntMdSP);
	ResetRCnt(RCntCNT2);

	dgx_cpu_time = 0, dgx_gpu_time = 0;

	printf("\nEngine version: %d.%d.%d%s\n", VERSION_ENGINE_MAJOR, VERSION_ENGINE_MINOR, VERSION_ENGINE_BUILD, VERSION_ENGINE_SUFFIX);
}

